"""
Custom hook to customize the behavior of Hatchling.
Here, we customize the tag of the generated wheels.
"""

import sysconfig
from typing import Any, Dict

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


def get_tag() -> str:
    """Get appropriate wheel tag according to system"""
    tag_platform = sysconfig.get_platform().replace("-", "_").replace(".", "_")
    return f"py3-none-{tag_platform}"


class CustomBuildHook(BuildHookInterface):
    """A custom build hook"""

    def initialize(self, version: str, build_data: Dict[str, Any]) -> None:
        """This step ccurs immediately before each build."""
        build_data["tag"] = get_tag()



"""Treelite Model builder class"""

import ctypes
import dataclasses
import json
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from .core import _LIB, _check_call
from .model import Model
from .util import c_array, c_str


@dataclasses.dataclass
class Metadata:
    """
    Metadata object, consisting of metadata information about the model at large.

    Parameters
    ----------
    num_feature:
        Number of features used in the model.
        We assume that all feature indices are between ``0`` and ``num_feature - 1``.
    task_type:
        Task type. Can be one of ``kBinaryClf``, ``kRegressor``, ``kMultiClf``,
        ``kLearningToRank``, or ``kIsolationForest``.
    average_tree_output:
        Whether to average outputs of trees
    num_target:
        Number of targets
    num_class:
        Number of classes. num_class[i] is the number of classes of target i.
    leaf_vector_shape:
        Shape of the output from each leaf node
    """

    num_feature: int
    task_type: str
    average_tree_output: bool
    num_target: int
    num_class: List[int]
    leaf_vector_shape: Tuple[int, int]

    def asdict(self) -> Dict[str, Any]:
        """Convert to dictionary"""
        return dataclasses.asdict(self)


@dataclasses.dataclass
class TreeAnnotation:
    """
    Annotation for individual trees. Use this object to look up which target and class
    each tree is associated with.

    The output of each target / class is obtained by summing the outputs of all trees that are
    associated with that target / class.
    target_id[i] indicates the target the i-th tree is associated with.
    (-1 indicates that the tree is a multi-target tree, whose output gets counted for all targets.)
    class_id[i] indicates the class the i-th tree is associated with.
    (-1 indicates that the tree's output gets counted for all classes.)

    Parameters
    ----------
    num_tree:
        Number of trees
    target_id:
        Target that each tree is associated with (see explanation above)
    class_id:
        Class that each tree is associated with (see explanation above)
    """

    num_tree: int
    target_id: List[int]
    class_id: List[int]

    def asdict(self) -> Dict[str, Any]:
        """Convert to dictionary"""
        return dataclasses.asdict(self)


@dataclasses.dataclass
class PostProcessorFunc:
    """
    Specification for postprocessor of prediction outputs

    Parameters
    ----------
    name:
        Name of the postprocessor. Consult :py:doc:`/knobs/postprocessor` for the list
        of available postprocessor functions.
    sigmoid_alpha:
        Scaling parameter for sigmoid function ``sigmoid(x) = 1 / (1 + exp(-alpha * x))``.
        This parameter is applicable only when ``name="sigmoid"`` or ``name="multiclass_ova"``.
        It must be strictly positive.
    ratio_c:
        Scaling parameter for exponential standard ratio transformation
        ``expstdratio(x) = exp2(-x / c)``.
        This parameter is applicable only when ``name="exponential_standard_ratio"``.
    """

    name: str
    sigmoid_alpha: float = 1.0
    ratio_c: float = 1.0

    def asdict(self) -> Dict[str, Any]:
        """Convert to dictionary"""
        return {
            "name": self.name,
            "config": {
                "sigmoid_alpha": self.sigmoid_alpha,
                "ratio_c": self.ratio_c,
            },
        }


class ModelBuilder:
    """
    Model builder class, to iteratively build a tree ensemble model.

    .. note::

        The model builder object must be only accessed by a single thread. To build
        multiple trees in parallel, create multiple builder objects and use model
        concatenation (:py:meth:`~treelite.Model.concatenate`).

    Parameters
    ----------
    threshold_type:
        Type of thresholds in the tree model
    leaf_output_type:
        Type of leaf outputs in the tree model
    metadata:
        Model metadata
    tree_annotation:
        Annotation for individual trees
    postprocessor:
        Postprocessor for prediction outputs
    base_scores:
        Baseline scores for targets and classes, before adding tree outputs.
        Also known as the intercept.
    attributes:
        Arbitrary JSON object, to be stored in the "attributes" field in the model object.
    """

    def __init__(
        self,
        *,
        threshold_type: str,
        leaf_output_type: str,
        metadata: Metadata,
        tree_annotation: TreeAnnotation,
        postprocessor: PostProcessorFunc,
        base_scores: List[float],
        attributes: Optional[Dict[Any, Any]] = None,
    ):  # pylint: disable=R0913
        self._handle = None

        handle = ctypes.c_void_p()
        metadata_obj = {
            "threshold_type": threshold_type,
            "leaf_output_type": leaf_output_type,
            "metadata": metadata.asdict(),
            "tree_annotation": tree_annotation.asdict(),
            "postprocessor": postprocessor.asdict(),
            "base_scores": base_scores,
        }
        if attributes is not None:
            metadata_obj["attributes"] = attributes
        metadata_str = json.dumps(metadata_obj)
        _check_call(
            _LIB.TreeliteGetModelBuilder(
                c_str(metadata_str),
                ctypes.byref(handle),
            )
        )
        self._handle = handle
        self.threshold_type = threshold_type
        self.leaf_output_type = leaf_output_type

    def __del__(self):
        if self.handle is not None:
            _check_call(_LIB.TreeliteDeleteModelBuilder(self._handle))
            self._handle = None

    @property
    def handle(self):
        """Access the handle to the associated C++ object"""
        return self._handle

    def start_tree(self):
        """Start a new tree"""
        _check_call(_LIB.TreeliteModelBuilderStartTree(self.handle))

    def end_tree(self):
        """End the current tree"""
        _check_call(_LIB.TreeliteModelBuilderEndTree(self.handle))

    def start_node(self, node_key: int):
        """
        Start a new node

        Parameters
        ----------
        node_key:
            Integer key that unique identifies the node
        """
        _check_call(
            _LIB.TreeliteModelBuilderStartNode(self.handle, ctypes.c_int(node_key))
        )

    def end_node(self):
        """End the current node"""
        _check_call(_LIB.TreeliteModelBuilderEndNode(self.handle))

    def numerical_test(
        self,
        feature_id: int,
        threshold: float,
        default_left: bool,
        opname: str,
        left_child_key: int,
        right_child_key: int,
    ):  # pylint: disable=R0913
        """
        Declare the current node as a numerical test node, where the test is of form
        [feature value] [op] [threshold]. Data points for which the test evaluates to True
        will be mapped to the left child node; all other data points (for which the test
        evaluates to False) will be mapped to the right child node.

        Parameters
        ----------
        feature_id:
            Feature ID
        threshold
            Threshold
        default_left
            Whether the missing value should be mapped to the left child
        opname
            Comparison operator
        left_child_key
            Integer key that unique identifies the left child node.
        right_child_key
            Integer key that unique identifies the right child node.
        """
        _check_call(
            _LIB.TreeliteModelBuilderNumericalTest(
                self.handle,
                ctypes.c_int32(feature_id),
                ctypes.c_double(threshold),
                ctypes.c_int(1 if default_left else 0),
                c_str(opname),
                ctypes.c_int(left_child_key),
                ctypes.c_int(right_child_key),
            )
        )

    def categorical_test(
        self,
        feature_id: int,
        default_left: bool,
        category_list: List[int],
        category_list_right_child: bool,
        left_child_key: int,
        right_child_key: int,
    ):  # pylint: disable=R0913
        """
        Declare the current node as a categorical test node, where the test is of form
        [feature value] \\in [category list].

        Parameters
        ----------
        feature_id:
            Feature ID
        default_left:
            Whether the missing value should be mapped to the left child
        category_list:
            List of categories to be tested for match
        category_list_right_child:
            Whether the data points for which the test evaluates to True should be mapped to the
            right child or the left child.
        left_child_key:
            Integer key that unique identifies the left child node.
        right_child_key:
            Integer key that unique identifies the right child node.
        """
        _check_call(
            _LIB.TreeliteModelBuilderCategoricalTest(
                self.handle,
                ctypes.c_int(feature_id),
                ctypes.c_bool(default_left),
                c_array(ctypes.c_uint32, category_list),
                ctypes.c_size_t(len(category_list)),
                ctypes.c_int(1 if category_list_right_child else 0),
                ctypes.c_int(left_child_key),
                ctypes.c_int(right_child_key),
            )
        )

    def leaf(self, leaf_value: Union[float, Sequence[float]]):
        """
        Declare the current node as a leaf node

        Parameters
        ----------
        leaf_value:
            Value of leaf output
        """
        if isinstance(leaf_value, Sequence):
            # leaf_value is a list
            if self.leaf_output_type == "float32":
                _check_call(
                    _LIB.TreeliteModelBuilderLeafVectorFloat32(
                        self.handle,
                        c_array(ctypes.c_float, leaf_value),
                        ctypes.c_size_t(len(leaf_value)),
                    )
                )
            elif self.leaf_output_type == "float64":
                _check_call(
                    _LIB.TreeliteModelBuilderLeafVectorFloat64(
                        self.handle,
                        c_array(ctypes.c_double, leaf_value),
                        ctypes.c_size_t(len(leaf_value)),
                    )
                )
        else:
            # leaf_value is a scalar
            _check_call(
                _LIB.TreeliteModelBuilderLeafScalar(
                    self.handle,
                    ctypes.c_double(leaf_value),
                )
            )

    def gain(self, gain: float):
        """
        Specify the gain (loss reduction) that's resulted from the current split.

        Parameters
        ----------
        gain:
            Gain (loss reduction)
        """
        _check_call(
            _LIB.TreeliteModelBuilderGain(
                self.handle,
                ctypes.c_double(gain),
            )
        )

    def data_count(self, data_count: int):
        """
        Specify the number of data points (samples) that are mapped to the current node.

        Parameters
        ----------
        data_count:
            Number of data points
        """
        _check_call(
            _LIB.TreeliteModelBuilderGain(
                self.handle,
                ctypes.c_uint64(data_count),
            )
        )

    def sum_hess(self, sum_hess: float):
        """
        Specify the weighted sample count or the sum of Hessians for the data points that
        are mapped to the current node.

        Parameters
        ----------
        sum_hess:
            Weighted sample count or the sum of Hessians
        """
        _check_call(
            _LIB.TreeliteModelBuilderSumHess(
                self.handle,
                ctypes.c_double(sum_hess),
            )
        )

    def commit(self) -> Model:
        """
        Conclude model building and obtain the final model object.

        Returns
        -------
        model : :py:class:`Model`
            Finished model
        """
        model_handle = ctypes.c_void_p()
        _check_call(
            _LIB.TreeliteModelBuilderCommitModel(
                self.handle,
                ctypes.byref(model_handle),
            )
        )
        return Model(handle=model_handle)



"""
Custom hook to customize path for libtreelite.so
"""


def get_custom_libpath():
    """
    Get custom path for libtreelite.so.
    If valid, must return a directory containing libtreelite.so.
    """
    return None



"""Treelite Model class"""

from __future__ import annotations

import ctypes
import pathlib
import platform
import warnings
from typing import Any, List, Optional, Union

import numpy as np

from . import compat
from .core import _LIB, _check_call
from .util import c_array, c_str, py_str


class Model:
    """
    Decision tree ensemble model

    Parameters
    ----------
    handle :
        Handle to C++ object
    """

    def __init__(self, *, handle: Optional[Any] = None):
        self._handle = handle

    def __del__(self):
        if self.handle is not None:
            _check_call(_LIB.TreeliteFreeModel(self._handle))
            self._handle = None

    @property
    def handle(self):
        """Access the handle to the associated C++ object"""
        return self._handle

    @property
    def num_tree(self) -> int:
        """Number of decision trees in the model"""
        if self.handle is None:
            raise AttributeError("Model not loaded yet")
        out = ctypes.c_size_t()
        _check_call(_LIB.TreeliteQueryNumTree(self.handle, ctypes.byref(out)))
        return out.value

    @property
    def num_feature(self) -> int:
        """Number of features used in the model"""
        if self.handle is None:
            raise AttributeError("Model not loaded yet")
        out = ctypes.c_size_t()
        _check_call(_LIB.TreeliteQueryNumFeature(self.handle, ctypes.byref(out)))
        return out.value

    @property
    def input_type(self) -> str:
        """Input type"""
        if self.handle is None:
            raise AttributeError("Model not loaded yet")
        out = ctypes.c_char_p()
        _check_call(_LIB.TreeliteGetInputType(self.handle, ctypes.byref(out)))
        return py_str(out.value)

    @property
    def output_type(self) -> str:
        """Output type"""
        if self.handle is None:
            raise AttributeError("Model not loaded yet")
        out = ctypes.c_char_p()
        _check_call(_LIB.TreeliteGetOutputType(self.handle, ctypes.byref(out)))
        return py_str(out.value)

    @classmethod
    def concatenate(cls, model_objs: List[Model]) -> Model:
        """
        Concatenate multiple model objects into a single model object by copying
        all member trees into the destination model object.

        Parameters
        ----------
        model_objs :
            List of :py:class:`Model` objects

        Returns
        -------
        model : :py:class:`Model`
            Concatenated model

        Example
        -------
        .. code-block:: python

           concatenated_model = treelite.Model.concatenate([model1, model2, model3])
        """
        concatenated_model_handle = ctypes.c_void_p()
        model_obj_handles = []
        for i, obj in enumerate(model_objs):
            if obj.handle is None:
                raise RuntimeError(f"Model at index {i} not loaded yet")
            model_obj_handles.append(obj.handle)
        _check_call(
            _LIB.TreeliteConcatenateModelObjects(
                c_array(ctypes.c_void_p, model_obj_handles),
                len(model_obj_handles),
                ctypes.byref(concatenated_model_handle),
            )
        )
        return Model(handle=concatenated_model_handle)

    @classmethod
    def load(
        cls, filename: str, model_format: str, allow_unknown_field: bool = False
    ) -> Model:
        """
        Deprecated; please use :py:meth:`~treelite.frontend.load_xgboost_model` instead.
        Load a tree ensemble model from a file.

        Parameters
        ----------
        filename :
            Path to model file
        model_format :
            Model file format. Must be "xgboost", "xgboost_json", or "lightgbm"
        allow_unknown_field:
            Whether to allow extra fields with unrecognized keys. This flag is only
            applicable if model_format="xgboost_json"

        Returns
        -------
        model : :py:class:`Model`
            Loaded model
        """
        model_format = model_format.lower()

        def deprecation_warning(alt: str):
            warnings.warn(
                (
                    "treelite.Model.load() is deprecated. "
                    f"Use treelite.frontend.{alt}() instead."
                ),
                FutureWarning,
            )

        if model_format == "xgboost":
            deprecation_warning("load_xgboost_model_legacy_binary")
            return Model(handle=compat.load_xgboost_model_legacy_binary(filename))
        if model_format == "xgboost_json":
            deprecation_warning("load_xgboost_model")
            return Model(
                handle=compat.load_xgboost_model(
                    filename, allow_unknown_field=allow_unknown_field
                )
            )
        if model_format == "lightgbm":
            deprecation_warning("load_lightgbm_model")
            return Model(handle=compat.load_lightgbm_model(filename))
        raise ValueError(f"Unknown model format {model_format}")

    @classmethod
    def from_xgboost(cls, booster: Any) -> Model:
        """
        Deprecated; please use :py:meth:`~treelite.frontend.from_xgboost` instead.
        Load a tree ensemble model from an XGBoost Booster object.

        Parameters
        ----------
        booster : Object of type :py:class:`xgboost.Booster`
            Python handle to XGBoost model

        Returns
        -------
        model : :py:class:`Model`
            Loaded model
        """
        warnings.warn(
            (
                "treelite.Model.from_xgboost() is deprecated. "
                "Use treelite.frontend.from_xgboost() instead."
            ),
            FutureWarning,
        )
        return Model(handle=compat.from_xgboost(booster))

    @classmethod
    def from_xgboost_json(
        cls,
        model_json_str: Union[bytes, bytearray, str],
        *,
        allow_unknown_field: bool = False,
    ) -> Model:
        """
        Deprecated; please use :py:meth:`~treelite.frontend.from_xgboost_json` instead.
        Load a tree ensemble model from a string containing XGBoost JSON.

        Parameters
        ----------
        model_json_str :
            A string specifying an XGBoost model in the XGBoost JSON format
        allow_unknown_field:
            Whether to allow extra fields with unrecognized keys

        Returns
        -------
        model : :py:class:`Model`
            Loaded model
        """
        warnings.warn(
            (
                "treelite.Model.from_xgboost_json() is deprecated. "
                "Use treelite.frontend.from_xgboost_json() instead."
            ),
            FutureWarning,
        )
        return Model(
            handle=compat.from_xgboost_json(
                model_json_str, allow_unknown_field=allow_unknown_field
            )
        )

    @classmethod
    def from_lightgbm(cls, booster):
        """
        Deprecated; please use :py:meth:`~treelite.frontend.from_lightgbm` instead.
        Load a tree ensemble model from a LightGBM Booster object.

        Parameters
        ----------
        booster : object of type :py:class:`lightgbm.Booster`
            Python handle to LightGBM model

        Returns
        -------
        model : :py:class:`Model`
            loaded model
        """
        warnings.warn(
            (
                "treelite.Model.from_lightgbm() is deprecated. "
                "Use treelite.frontend.from_lightgbm() instead."
            ),
            FutureWarning,
        )
        return Model(handle=compat.from_lightgbm(booster))

    def dump_as_json(self, *, pretty_print: bool = True) -> str:
        """
        Dump the model as a JSON string. This is useful for inspecting details of the tree
        ensemble model.

        Parameters
        ----------
        pretty_print :
            Whether to pretty-print the JSON string, set this to False to make the string compact

        Returns
        -------
        json_str : str
            JSON string representing the model
        """
        json_str = ctypes.c_char_p()
        _check_call(
            _LIB.TreeliteDumpAsJSON(
                self.handle,
                ctypes.c_int(1 if pretty_print else 0),
                ctypes.byref(json_str),
            )
        )
        return py_str(json_str.value)

    def get_header_accessor(self) -> HeaderAccessor:
        """
        Obtain accessor for fields in the header.
        See :ref:`field_accessors` for more details.
        """
        return HeaderAccessor(self)

    def get_tree_accessor(self, tree_id: int) -> TreeAccessor:
        """
        Obtain accessor for fields in a tree.
        See :ref:`field_accessors` for more details.

        Parameters
        ----------
        tree_id:
            ID of the tree
        """
        return TreeAccessor(self, tree_id=tree_id)

    def serialize(self, filename: Union[str, pathlib.Path]):
        """
        Serialize (persist) the model to a checkpoint file in the disk, using a fast binary
        representation. To recover the model from the checkpoint, use :py:func:`deserialize`
        method.

        .. note:: Notes on forward and backward compatibility

            Please see :doc:`/serialization/index`.

        Parameters
        ----------
        filename :
            Path to checkpoint
        """
        filepath = pathlib.Path(filename).expanduser().resolve()
        _check_call(
            _LIB.TreeliteSerializeModelToFile(self.handle, c_str(str(filepath)))
        )

    def serialize_bytes(self) -> bytes:
        """
        Serialize (persist) the model to a byte sequence, using a fast binary representation.
        To recover the model from the byte sequence, use :py:func:`deserialize_bytes` method.

        .. note:: Notes on forward and backward compatibility

            Please see :doc:`/serialization/index`.
        """
        char_ptr_t = ctypes.POINTER(ctypes.c_char)
        out_bytes = char_ptr_t()
        out_bytes_len = ctypes.c_size_t()
        _check_call(
            _LIB.TreeliteSerializeModelToBytes(
                self.handle, ctypes.byref(out_bytes), ctypes.byref(out_bytes_len)
            )
        )
        return ctypes.string_at(out_bytes, out_bytes_len.value)

    @classmethod
    def deserialize(cls, filename: Union[str, pathlib.Path]) -> Model:
        """
        Deserialize (recover) the model from a checkpoint file in the disk. It is expected that
        the file was generated by a call to the :py:func:`serialize` method.

        .. note:: Notes on forward and backward compatibility

            Please see :doc:`/serialization/index`.

        Parameters
        ----------
        filename :
            Path to checkpoint

        Returns
        -------
        model : :py:class:`Model`
            Recovered model
        """
        handle = ctypes.c_void_p()
        filepath = pathlib.Path(filename).expanduser().resolve()
        _check_call(
            _LIB.TreeliteDeserializeModelFromFile(
                c_str(str(filepath)), ctypes.byref(handle)
            )
        )
        return Model(handle=handle)

    @classmethod
    def deserialize_bytes(cls, model_bytes: bytes) -> Model:
        """
        Deserialize (recover) the model from a byte sequence. It is expected that
        the byte sequence was generated by a call to the :py:func:`serialize_bytes` method.

        .. note:: Notes on forward and backward compatibility

            Please see :doc:`/serialization/index`.

        Parameters
        ----------
        model_bytes :
            Byte sequence representing the serialized model

        Returns
        -------
        model : :py:class:`Model`
            Recovered model
        """
        handle = ctypes.c_void_p()
        model_bytes_len = len(model_bytes)
        buffer = ctypes.create_string_buffer(model_bytes, model_bytes_len)
        _check_call(
            _LIB.TreeliteDeserializeModelFromBytes(
                ctypes.pointer(buffer),
                ctypes.c_size_t(model_bytes_len),
                ctypes.byref(handle),
            )
        )
        return Model(handle=handle)


class _TreelitePyBufferFrame(ctypes.Structure):  # pylint: disable=R0903
    """Abridged buffer structure used by Treelite"""

    _fields_ = [
        ("buf", ctypes.c_void_p),
        ("format", ctypes.c_char_p),
        ("itemsize", ctypes.c_size_t),
        ("nitem", ctypes.c_size_t),
    ]


class _PyBuffer(ctypes.Structure):  # pylint: disable=R0902,R0903
    """The full Python buffer structure as defined by PEP 3118"""

    _fields_ = (
        ("buf", ctypes.c_void_p),
        ("obj", ctypes.py_object),
        ("len", ctypes.c_ssize_t),
        ("itemsize", ctypes.c_ssize_t),
        ("readonly", ctypes.c_int),
        ("ndim", ctypes.c_int),
        ("format", ctypes.c_char_p),
        ("shape", ctypes.POINTER(ctypes.c_ssize_t)),
        ("strides", ctypes.POINTER(ctypes.c_ssize_t)),
        ("suboffsets", ctypes.POINTER(ctypes.c_ssize_t)),
        ("internal", ctypes.c_void_p),
    )


if platform.python_implementation() == "CPython":
    ctypes.pythonapi.PyMemoryView_FromBuffer.argtypes = [ctypes.POINTER(_PyBuffer)]
    ctypes.pythonapi.PyMemoryView_FromBuffer.restype = ctypes.py_object
    ctypes.pythonapi.PyObject_GetBuffer.argtypes = [
        ctypes.py_object,
        ctypes.POINTER(_PyBuffer),
        ctypes.c_int,
    ]
    ctypes.pythonapi.PyObject_GetBuffer.restype = ctypes.c_int


def _pybuffer2numpy(frame: _TreelitePyBufferFrame) -> np.ndarray:
    if platform.python_implementation() != "CPython":
        raise NotImplementedError("_pybuffer2numpy() not supported on PyPy")
    if not frame.buf:
        if frame.format == b"=l":
            dtype = "int32"
        elif frame.format == b"=Q":
            dtype = "uint64"
        elif frame.format == b"=L":
            dtype = "uint32"
        elif frame.format == b"=B":
            dtype = "uint8"
        elif frame.format == b"=f":
            dtype = "float32"
        elif frame.format == b"=d":
            dtype = "float64"
        else:
            raise RuntimeError(
                f"Unrecognized format string: {frame.format.decode('utf-8')}"
            )
        return np.array([], dtype=dtype)
    py_buf = _PyBuffer()
    py_buf.buf = frame.buf
    py_buf.obj = ctypes.py_object(frame)
    py_buf.len = frame.nitem * frame.itemsize
    py_buf.itemsize = frame.itemsize
    py_buf.readonly = 0
    py_buf.ndim = 1
    py_buf.format = frame.format
    py_buf.shape = (ctypes.c_ssize_t * 1)(frame.nitem)
    py_buf.strides = (ctypes.c_ssize_t * 1)(frame.itemsize)
    py_buf.suboffsets = None
    py_buf.internal = None

    view: memoryview = ctypes.pythonapi.PyMemoryView_FromBuffer(ctypes.byref(py_buf))
    return np.asarray(view)


def _numpy2pybuffer(array: np.ndarray) -> _TreelitePyBufferFrame:
    if platform.python_implementation() != "CPython":
        raise NotImplementedError("_numpy2pybuffer() not supported on PyPy")
    if len(array.shape) != 1:
        raise ValueError("Cannot handle NumPy array that has more than 1 dimension")
    view: memoryview = array.data
    buffer = _PyBuffer()
    if (
        ctypes.pythonapi.PyObject_GetBuffer(
            ctypes.py_object(view),
            ctypes.byref(buffer),
            ctypes.c_int(0),  # PyBUF_SIMPLE
        )
        != 0
    ):
        raise RuntimeError("Call to PyObject_GetBuffer() failed")
    frame = _TreelitePyBufferFrame()
    frame.buf = buffer.buf
    frame.format = buffer.format
    frame.itemsize = buffer.itemsize
    frame.nitem = array.shape[0]
    return frame


class HeaderAccessor:
    """
    Accessor for fields in the header

    Parameters
    ----------
    model:
        The model object
    """

    def __init__(self, model: Model):
        self._model = model

    def get_field(self, name: str) -> Union[np.ndarray, str]:
        """
        Get a field

        Parameters
        ----------
        name:
            Name of the field. Consult :doc:`the model spec </serialization/v4>`
            for the list of fields.

        Returns
        -------
        field: :py:class:`numpy.ndarray` or :py:class:`str`
            Value in the field
            (``str`` for a string field, ``np.ndarray`` for other fields)
        """
        if platform.python_implementation() != "CPython":
            raise NotImplementedError("get_field() not supported on PyPy")
        obj = _TreelitePyBufferFrame()
        _check_call(
            _LIB.TreeliteGetHeaderField(
                self._model.handle,
                c_str(name),
                ctypes.byref(obj),
            )
        )
        array = _pybuffer2numpy(obj)
        if array.dtype == "S1":
            return array.tobytes().decode("utf-8")
        return array

    def set_field(self, name: str, value: Union[np.ndarray, str]):
        """
        Set a field

        Parameters
        ----------
        name:
            Name of the field. Consult :doc:`the model spec </serialization/v4>`
            for the list of fields.
        value:
            New value for the field
            (``str`` for a string field, ``np.ndarray`` for other fields)
        """
        if platform.python_implementation() != "CPython":
            raise NotImplementedError("set_field() not supported on PyPy")
        if isinstance(value, str):
            value = np.frombuffer(value.encode("utf-8"), dtype="S1")
        _check_call(
            _LIB.TreeliteSetHeaderField(
                self._model.handle,
                c_str(name),
                _numpy2pybuffer(value),
            )
        )


class TreeAccessor:
    """
    Accessor for fields in a tree

    Parameters
    ----------
    model:
        The model object
    tree_id:
        ID of the tree
    """

    def __init__(self, model: Model, *, tree_id: int):
        self._model = model
        self._tree_id = tree_id

    def get_field(self, name: str) -> np.ndarray:
        """
        Get a field

        Parameters
        ----------
        name:
            Name of the field. Consult :doc:`the model spec </serialization/v4>`
            for the list of fields.

        Returns
        -------
        field: :py:class:`numpy.ndarray`
            Value in the field
        """
        obj = _TreelitePyBufferFrame()
        _check_call(
            _LIB.TreeliteGetTreeField(
                self._model.handle,
                ctypes.c_uint64(self._tree_id),
                c_str(name),
                ctypes.byref(obj),
            )
        )
        return _pybuffer2numpy(obj)

    def set_field(self, name: str, value: np.ndarray):
        """
        Set a field

        Parameters
        ----------
        name:
            Name of the field. Consult :doc:`the model spec </serialization/v4>`
            for the list of fields.
        value:
            New value for the field
        """
        _check_call(
            _LIB.TreeliteSetTreeField(
                self._model.handle,
                ctypes.c_uint64(self._tree_id),
                c_str(name),
                _numpy2pybuffer(value),
            )
        )


__all__ = ["Model", "HeaderAccessor", "TreeAccessor"]



"""
Functions for building libtreelite
"""

import logging
import os
import pathlib
import shutil
import subprocess
import sys
from platform import system
from typing import Optional

from .build_config import BuildConfiguration


def _lib_name() -> str:
    """Return platform dependent shared object name."""
    if system() in ["Linux", "OS400"] or system().upper().endswith("BSD"):
        name = "libtreelite.so"
    elif system() == "Darwin":
        name = "libtreelite.dylib"
    elif system() == "Windows":
        name = "treelite.dll"
    else:
        raise NotImplementedError(f"System {system()} not supported")
    return name


def build_libtreelite(
    cpp_src_dir: pathlib.Path,
    build_dir: pathlib.Path,
    build_config: BuildConfiguration,
) -> pathlib.Path:
    """Build libtreelite in a temporary directory and obtain the path to built libtreelite"""
    logger = logging.getLogger("treelite.packager.build_libtreelite")

    if not cpp_src_dir.is_dir():
        raise RuntimeError(f"Expected {cpp_src_dir} to be a directory")
    logger.info(
        "Building %s from the C++ source files in %s...", _lib_name(), str(cpp_src_dir)
    )

    def _build(*, generator: str) -> None:
        cmake_cmd = [
            "cmake",
            str(cpp_src_dir),
            generator,
        ]
        cmake_cmd.extend(build_config.get_cmake_args())

        # Flag for cross-compiling for Apple Silicon
        # We use environment variable because it's the only way to pass down custom flags
        # through the cibuildwheel package, which calls `pip wheel` command.
        if "CIBW_TARGET_OSX_ARM64" in os.environ:
            cmake_cmd.extend(
                ["-DCMAKE_OSX_ARCHITECTURES=arm64", "-DDETECT_CONDA_ENV=OFF"]
            )

        logger.info("CMake args: %s", str(cmake_cmd))
        subprocess.check_call(cmake_cmd, cwd=build_dir)

        if system() == "Windows":
            subprocess.check_call(
                ["cmake", "--build", ".", "--config", "Release"], cwd=build_dir
            )
        else:
            nproc = os.cpu_count()
            assert build_tool is not None
            subprocess.check_call([build_tool, f"-j{nproc}"], cwd=build_dir)

    if system() == "Windows":
        supported_generators = (
            "-GVisual Studio 17 2022",
            "-GVisual Studio 16 2019",
            "-GVisual Studio 15 2017",
            "-GMinGW Makefiles",
        )
        for generator in supported_generators:
            try:
                _build(generator=generator)
                logger.info(
                    "Successfully built %s using generator %s", _lib_name(), generator
                )
                break
            except subprocess.CalledProcessError as e:
                logger.info(
                    "Tried building with generator %s but failed with exception %s",
                    generator,
                    str(e),
                )
                # Empty build directory
                shutil.rmtree(build_dir)
                build_dir.mkdir()
        else:
            raise RuntimeError(
                "None of the supported generators produced a successful build!"
                f"Supported generators: {supported_generators}"
            )
    else:
        build_tool = "ninja" if shutil.which("ninja") else "make"
        generator = "-GNinja" if build_tool == "ninja" else "-GUnix Makefiles"
        try:
            _build(generator=generator)
        except subprocess.CalledProcessError as e:
            logger.info("Failed to build with OpenMP. Exception: %s", str(e))
            build_config.use_openmp = False
            _build(generator=generator)

    return build_dir / _lib_name()


def locate_local_libtreelite(
    toplevel_dir: pathlib.Path,
    logger: logging.Logger,
) -> Optional[pathlib.Path]:
    """
    Locate libtreelite from the local project directory's lib/ subdirectory.
    """
    libtreelite = toplevel_dir.parent / "build" / _lib_name()
    if libtreelite.exists():
        logger.info("Found %s at %s", libtreelite.name, str(libtreelite.parent))
        return libtreelite
    logger.info("Did not find %s at %s", libtreelite.name, str(libtreelite.parent))
    return None


def locate_or_build_libtreelite(
    toplevel_dir: pathlib.Path,
    build_dir: pathlib.Path,
    build_config: BuildConfiguration,
) -> pathlib.Path:
    """Locate libtreelite; if not exist, build it"""
    logger = logging.getLogger("treelite.packager.locate_or_build_libtreelite")

    if build_config.use_system_libtreelite:
        # If the user explicitly specifies the path for libtreelite, use it
        if build_config.system_libtreelite_dir:
            p = pathlib.Path(build_config.system_libtreelite_dir)
            logger.info(
                "system_libtreelite_dir was specified. Locating %s from path %s...",
                _lib_name(),
                str(p),
            )
            sys_prefix_candidates = [p.expanduser().resolve()]
        else:
            # Find libtreelite from system prefix
            sys_prefix = pathlib.Path(sys.base_prefix)
            sys_prefix_candidates = [
                sys_prefix / "lib",
                # Paths possibly used on Windows
                sys_prefix / "bin",
                sys_prefix / "Library",
                sys_prefix / "Library" / "bin",
                sys_prefix / "Library" / "lib",
            ]
            sys_prefix_candidates = [
                p.expanduser().resolve() for p in sys_prefix_candidates
            ]
        for candidate_dir in sys_prefix_candidates:
            libtreelite_sys = candidate_dir / _lib_name()
            if libtreelite_sys.exists():
                logger.info("Using system treelite: %s", str(libtreelite_sys))
                return libtreelite_sys
        rec_msg = (
            "Make sure that system_libtreelite_dir is a valid path"
            if build_config.system_libtreelite_dir
            else "Consider setting system_libtreelite_dir in the build configuration"
        )
        raise RuntimeError(
            f"use_system_libtreelite was specified but {_lib_name()} is not found. {rec_msg}. "
            "Paths searched (in order): \n"
            + "\n".join([f"* {str(p)}" for p in sys_prefix_candidates])
        )

    libtreelite = locate_local_libtreelite(toplevel_dir, logger=logger)
    if libtreelite is not None:
        return libtreelite

    if toplevel_dir.joinpath("cpp_src").exists():
        # Source distribution; all C++ source files to be found in cpp_src/
        cpp_src_dir = toplevel_dir.joinpath("cpp_src")
    else:
        # Probably running "pip install ." from python-package/
        cpp_src_dir = toplevel_dir.parent
        if not cpp_src_dir.joinpath("CMakeLists.txt").exists():
            raise RuntimeError(f"Did not find CMakeLists.txt from {cpp_src_dir}")
    return build_libtreelite(
        cpp_src_dir, build_dir=build_dir, build_config=build_config
    )



"""Treelite Model builder class (legacy)"""

# pylint: disable=E1135, E1137

from __future__ import annotations

import dataclasses
import warnings
from typing import Dict, List, Optional, Union

from .core import TreeliteError
from .model import Model
from .model_builder import Metadata
from .model_builder import ModelBuilder as ModelBuilderNew
from .model_builder import PostProcessorFunc, TreeAnnotation


@dataclasses.dataclass
class _LeafNode:
    leaf_value: Union[float, List[float]]


@dataclasses.dataclass
class _NumericalTestNode:
    feature_id: int
    opname: str
    threshold: float
    default_left: bool
    left_child_key: int
    right_child_key: int


@dataclasses.dataclass
class _CategoricalTestNode:
    feature_id: int
    left_categories: List[int]
    default_left: bool
    left_child_key: int
    right_child_key: int


class ModelBuilder:
    """
    Legacy model builder class. **New code should use the new model builder**
    :py:class:`treelite.model_builder.ModelBuilder` **instead.**

    This module is meant to enable existing code using the old model builder API
    to continue functioning. Users are highly encouraged to migrate to the new
    model builder API, to take advantage of new features including the support
    for multi-target tree models.

    Parameters
    ----------
    num_feature :
        Number of features used in model being built. We assume that all
        feature indices are between ``0`` and ``num_feature - 1``.
    num_class :
        Number of output groups; ``>1`` indicates multiclass classification
    average_tree_output :
        Whether the model is a random forest; ``True`` indicates a random forest
        and ``False`` indicates gradient boosted trees
    threshold_type:
        Type of thresholds in the tree model
    leaf_output_type:
        Type of leaf outputs in the tree model
    pred_transform:
        Postprocessor for prediction outputs
    sigmoid_alpha:
        Scaling parameter for sigmoid function ``sigmoid(x) = 1 / (1 + exp(-alpha * x))``.
        This parameter is applicable only when ``name="sigmoid"`` or ``name="multiclass_ova"``.
        It must be strictly positive.
    ratio_c:
        Scaling parameter for exponential standard ratio transformation
        ``expstdratio(x) = exp2(-x / c)``.
        This parameter is applicable only when ``name="exponential_standard_ratio"``.
    global_bias:
        Global bias of the model.
        Predicted margin scores of all instances will be  adjusted by the global bias.
    """

    class Node:
        """A node in a tree"""

        def __init__(self):
            self.empty = True
            self.node_key: Optional[int] = None
            self.tree: Optional[ModelBuilder.Tree] = None
            self.node: Optional[
                Union[_LeafNode, _NumericalTestNode, _CategoricalTestNode]
            ] = None

        def __repr__(self):
            return "<treelite.ModelBuilder.Node object>"

        def set_root(self):
            """Set the node as the root"""
            if self.tree:
                self.tree._set_root(node_key=self.node_key)  # pylint: disable=W0212
            else:
                raise TreeliteError(
                    "This node has never been inserted into a tree; "
                    "a node must be inserted before it can be a root"
                )

        def set_leaf_node(
            self,
            leaf_value: Union[float, List[float]],
            leaf_value_type: str = "float32",
        ):
            """
            Set the node as a leaf node

            Parameters
            ----------
            leaf_value :
                Usually a single leaf value (weight) of the leaf node. For multi-class
                random forest classifier, leaf_value should be a list of leaf weights.
            leaf_value_type :
                Data type used for leaf_value
            """

            if not self.empty:
                raise ValueError(
                    "Cannot modify a non-empty node. "
                    "If you meant to change type of the node, "
                    "delete it first and then add an empty node with "
                    "the same key."
                )
            if self.tree is None:
                raise TreeliteError(
                    "This node has never been inserted into a tree; "
                    "a node must be inserted before it can be a leaf node"
                )
            if self.tree.leaf_output_type != leaf_value_type:
                raise ValueError(
                    "Leaf value must have same type as the tree's leaf_output_type "
                    f"({self.tree.leaf_output_type})."
                )
            self.node = _LeafNode(leaf_value=leaf_value)

        # pylint: disable=R0913
        def set_numerical_test_node(
            self,
            feature_id: int,
            opname: str,
            threshold: float,
            default_left: bool,
            left_child_key: int,
            right_child_key: int,
            threshold_type: str = "float32",
        ):
            """
            Set the node as a test node with numerical split. The test is in the form
            ``[feature value] OP [threshold]``. Depending on the result of the test,
            either left or right child would be taken.

            Parameters
            ----------
            feature_id :
                Feature index
            opname :
                Binary operator to use in the test
            threshold :
                Threshold value
            default_left :
                Default direction for missing values
                (``True`` for left; ``False`` for right)
            left_child_key :
                Unique integer key to identify the left child node
            right_child_key :
                Unique integer key to identify the right child node
            threshold_type :
                Data type for threshold value (e.g. 'float32')
            """
            if not self.empty:
                raise ValueError(
                    "Cannot modify a non-empty node. "
                    "If you meant to change type of the node, "
                    "delete it first and then add an empty node with "
                    "the same key."
                )
            if self.tree is None:
                raise TreeliteError(
                    "This node has never been inserted into a tree; "
                    "a node must be inserted before it can be a test node"
                )
            # Automatically create child nodes that don't exist yet
            if left_child_key not in self.tree:
                self.tree[left_child_key] = ModelBuilder.Node()
            if right_child_key not in self.tree:
                self.tree[right_child_key] = ModelBuilder.Node()
            if self.tree.threshold_type != threshold_type:
                raise ValueError(
                    "Threshold value must have same type as the tree's threshold_type "
                    f"({self.tree.threshold_type})."
                )
            self.node = _NumericalTestNode(
                feature_id=feature_id,
                opname=opname,
                threshold=threshold,
                default_left=default_left,
                left_child_key=left_child_key,
                right_child_key=right_child_key,
            )
            self.empty = False

        # pylint: disable=R0913
        def set_categorical_test_node(
            self,
            feature_id: int,
            left_categories: List[int],
            default_left: bool,
            left_child_key: int,
            right_child_key: int,
        ):
            """
            Set the node as a test node with categorical split. A list defines all
            categories that would be classified as the left side. Categories are
            integers ranging from ``0`` to ``n-1``, where ``n`` is the number of
            categories in that particular feature.

            Parameters
            ----------
            feature_id :
                Feature index
            left_categories :
                List of categories belonging to the left child.
            default_left :
                Default direction for missing values
                (``True`` for left; ``False`` for right)
            left_child_key :
                Unique integer key to identify the left child node
            right_child_key :
                Unique integer key to identify the right child node
            """
            if not self.empty:
                raise ValueError(
                    "Cannot modify a non-empty node. "
                    "If you meant to change type of the node, "
                    "delete it first and then add an empty node with "
                    "the same key."
                )
            if self.tree is None:
                raise TreeliteError(
                    "This node has never been inserted into a tree; "
                    "a node must be inserted before it can be a test node"
                )
            # Automatically create child nodes that don't exist yet
            if left_child_key not in self.tree:
                self.tree[left_child_key] = ModelBuilder.Node()
            if right_child_key not in self.tree:
                self.tree[right_child_key] = ModelBuilder.Node()
            self.node = _CategoricalTestNode(
                feature_id=feature_id,
                left_categories=left_categories,
                default_left=default_left,
                left_child_key=left_child_key,
                right_child_key=right_child_key,
            )
            self.empty = False

    class Tree:
        """
        A decision tree in a tree ensemble builder

        Parameters
        ----------
        threshold_type:
            Type of thresholds in the tree model
        leaf_output_type:
            Type of leaf outputs in the tree model
        """

        def __init__(
            self,
            threshold_type: str = "float32",
            leaf_output_type: str = "float32",
        ):
            self.ensemble: Optional[ModelBuilder] = None
            self.nodes: Dict[int, ModelBuilder.Node] = {}
            self.threshold_type = threshold_type
            self.leaf_output_type = leaf_output_type

        # Implement dict semantics whenever applicable
        def items(self):  # pylint: disable=C0111
            return self.nodes.items()

        def keys(self):  # pylint: disable=C0111
            return self.nodes.keys()

        def values(self):  # pylint: disable=C0111
            return self.nodes.values()

        def __len__(self):
            return len(self.nodes)

        def __getitem__(self, key):
            if key not in self.nodes:
                # Implicitly create a new node
                self.__setitem__(key, ModelBuilder.Node())
            return self.nodes.__getitem__(key)

        def __setitem__(self, key, value):
            if not isinstance(value, ModelBuilder.Node):
                raise ValueError("Value must be of type ModelBuidler.Node")
            if key in self.nodes:
                raise KeyError(
                    "Nodes with duplicate keys are not allowed. "
                    f"If you meant to replace node {key}, "
                    "delete it first and then add an empty node with "
                    "the same key."
                )
            if not value.empty:
                raise ValueError("Can only insert an empty node")
            self.nodes.__setitem__(key, value)
            value.node_key = key  # Save node id for later
            value.tree = self

        def __delitem__(self, key):
            self.nodes.__delitem__(key)

        def __iter__(self):
            return self.nodes.__iter__()

        def __repr__(self):
            return f"<treelite.ModelBuilder.Tree object containing {len(self.nodes)} nodes>\n"

        def __contains__(self, item):
            return self.nodes.__contains__(item)

        def _set_root(self, node_key: Optional[int]):
            assert node_key is not None
            self.root_key = node_key

    def __init__(
        self,
        num_feature: int,
        num_class: int = 1,
        average_tree_output: bool = False,
        threshold_type: str = "float32",
        leaf_output_type: str = "float32",
        *,
        pred_transform: str = "identity",
        sigmoid_alpha: float = 1.0,
        ratio_c: float = 1.0,
        global_bias: float = 0.0,
    ):  # pylint: disable=R0913
        warnings.warn(
            "treelite.ModelBuilder is deprecated and will be removed in Treelite 4.1. "
            "Please use treelite.model_builder.ModelBuilder class instead. The new model builder "
            "supports many more features such as support for multi-target tree models.",
            FutureWarning,
        )
        if num_class == 1:
            if pred_transform == "sigmoid":
                task_type = "kBinaryClf"
            else:
                task_type = "kRegressor"
            grove_per_class = False
        elif num_class > 1:
            task_type = "kMultiClf"
            grove_per_class = True
        else:
            raise ValueError("num_class must be at least 1")
        if pred_transform == "max_index":
            # max_index not supported; replace with softmax
            pred_transform = "softmax"
        if threshold_type == "unit32" or leaf_output_type == "unit32":
            raise ValueError(
                "Integer type (uint32) is no longer supported."
                "Please use float32 or float64 for thresholds and leaf outputs."
            )
        self.trees: List[ModelBuilder.Tree] = []
        self.threshold_type = threshold_type
        self.leaf_output_type = leaf_output_type
        self.metadata = Metadata(
            num_feature=num_feature,
            task_type=task_type,
            average_tree_output=average_tree_output,
            num_target=1,
            num_class=[num_class],
            leaf_vector_shape=(1, 1),
        )
        self.grove_per_class = grove_per_class
        self.postprocessor = PostProcessorFunc(
            name=pred_transform,
            sigmoid_alpha=sigmoid_alpha,
            ratio_c=ratio_c,
        )
        self.base_scores = [global_bias]

    def insert(self, index: int, tree: ModelBuilder.Tree):
        """
        Insert a tree at specified location in the ensemble

        Parameters
        ----------
        index :
            Index of the element before which to insert the tree
        tree :
            Tree to be inserted
        """
        if not isinstance(index, int):
            raise ValueError("index must be of int type")
        if index < 0 or index > len(self):
            raise ValueError("index out of bounds")
        if not isinstance(tree, ModelBuilder.Tree):
            raise ValueError("tree must be of type ModelBuilder.Tree")
        tree.ensemble = self
        self.trees.insert(index, tree)

    def append(self, tree: ModelBuilder.Tree):
        """
        Add a tree at the end of the ensemble

        Parameters
        ----------
        tree :
            tree to be added
        """
        if not isinstance(tree, ModelBuilder.Tree):
            raise ValueError("tree must be of type ModelBuilder.Tree")
        self.insert(len(self), tree)

    def commit(self) -> Model:
        """
        Finalize the ensemble model

        Returns
        -------
        model : :py:class:`Model`
            Finished model
        """
        num_tree = len(self.trees)
        if self.grove_per_class:
            tree_annotation = TreeAnnotation(
                num_tree=len(self.trees),
                target_id=[0] * num_tree,
                class_id=[(i % self.metadata.num_class[0]) for i in range(num_tree)],
            )
        else:
            tree_annotation = TreeAnnotation(
                num_tree=len(self.trees),
                target_id=[0] * num_tree,
                class_id=[0] * num_tree,
            )
        builder = ModelBuilderNew(
            threshold_type=self.threshold_type,
            leaf_output_type=self.leaf_output_type,
            metadata=self.metadata,
            tree_annotation=tree_annotation,
            postprocessor=self.postprocessor,
            base_scores=self.base_scores,
        )
        for tree in self.trees:
            builder.start_tree()
            root_node = tree[tree.root_key].node
            del tree.nodes[tree.root_key]
            nodes = [(tree.root_key, root_node)] + [
                (k, v.node) for k, v in tree.nodes.items()
            ]
            for key, node in nodes:
                builder.start_node(key)
                if isinstance(node, _LeafNode):
                    # Leaf root
                    builder.leaf(node.leaf_value)
                elif isinstance(node, _NumericalTestNode):
                    builder.numerical_test(
                        feature_id=node.feature_id,
                        threshold=node.threshold,
                        default_left=node.default_left,
                        opname=node.opname,
                        left_child_key=node.left_child_key,
                        right_child_key=node.right_child_key,
                    )
                elif isinstance(node, _CategoricalTestNode):
                    builder.categorical_test(
                        feature_id=node.feature_id,
                        default_left=node.default_left,
                        category_list=node.left_categories,
                        category_list_right_child=False,
                        left_child_key=node.left_child_key,
                        right_child_key=node.right_child_key,
                    )
                else:
                    raise ValueError("Root node cannot be empty")
                builder.end_node()
            builder.end_tree()
        return builder.commit()

    # Implement list semantics whenever applicable
    def __len__(self):
        return len(self.trees)

    def __getitem__(self, index):
        return self.trees.__getitem__(index)

    def __delitem__(self, index):
        self.trees.__delitem__(index)

    def __iter__(self):
        return self.trees.__iter__()

    def __reversed__(self):
        return self.trees.__reversed__()

    def __repr__(self):
        return (
            f"<treelite.ModelBuilder object storing {len(self.trees)} decision trees>\n"
        )



"""
Functions for building sdist
"""

import logging
import pathlib

from .util import copy_with_logging, copytree_with_logging


def copy_cpp_src_tree(
    cpp_src_dir: pathlib.Path, target_dir: pathlib.Path, logger: logging.Logger
) -> None:
    """Copy C++ source tree into build directory"""

    for subdir in [
        "src",
        "include",
        "cmake",
    ]:
        copytree_with_logging(cpp_src_dir / subdir, target_dir / subdir, logger=logger)

    for filename in ["CMakeLists.txt", "LICENSE"]:
        copy_with_logging(cpp_src_dir.joinpath(filename), target_dir, logger=logger)



"""
Custom build backend for Treelite Python package.
Builds source distribution and binary wheels, following PEP 517 / PEP 660.
Reuses components of Hatchling (https://github.com/pypa/hatch/tree/master/backend) for the sake
of brevity.
"""

import dataclasses
import logging
import os
import pathlib
import tempfile
from contextlib import contextmanager
from typing import Any, Dict, Iterator, Optional, Union

import hatchling.build

from .build_config import BuildConfiguration
from .nativelib import locate_local_libtreelite, locate_or_build_libtreelite
from .sdist import copy_cpp_src_tree
from .util import copy_with_logging, copytree_with_logging


@contextmanager
def cd(path: Union[str, pathlib.Path]) -> Iterator[str]:  # pylint: disable=C0103
    """
    Temporarily change working directory.
    TODO(hcho3): Remove this once we adopt Python 3.11, which implements contextlib.chdir.
    """
    path = str(path)
    path = os.path.realpath(path)
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(cwd)


TOPLEVEL_DIR = pathlib.Path(__file__).parent.parent.absolute().resolve()
logging.basicConfig(level=logging.INFO)


# Aliases
get_requires_for_build_sdist = hatchling.build.get_requires_for_build_sdist
get_requires_for_build_wheel = hatchling.build.get_requires_for_build_wheel
get_requires_for_build_editable = hatchling.build.get_requires_for_build_editable


def build_wheel(
    wheel_directory: str,
    config_settings: Optional[Dict[str, Any]] = None,
    metadata_directory: Optional[str] = None,
) -> str:
    """Build a wheel"""
    logger = logging.getLogger("treelite.packager.build_wheel")

    build_config = BuildConfiguration()
    build_config.update(config_settings)
    logger.info("Parsed build configuration: %s", dataclasses.asdict(build_config))

    # Create tempdir with Python package + libtreelite
    with tempfile.TemporaryDirectory() as td:
        td_path = pathlib.Path(td)
        build_dir = td_path / "libbuild"
        build_dir.mkdir()

        workspace = td_path / "whl_workspace"
        workspace.mkdir()
        logger.info("Copying project files to temporary directory %s", str(workspace))

        copy_with_logging(TOPLEVEL_DIR / "pyproject.toml", workspace, logger=logger)
        copy_with_logging(TOPLEVEL_DIR / "hatch_build.py", workspace, logger=logger)
        copy_with_logging(TOPLEVEL_DIR / "README.rst", workspace, logger=logger)

        pkg_path = workspace / "treelite"
        copytree_with_logging(TOPLEVEL_DIR / "treelite", pkg_path, logger=logger)
        lib_path = pkg_path / "lib"
        lib_path.mkdir()
        libtreelite = locate_or_build_libtreelite(
            TOPLEVEL_DIR, build_dir=build_dir, build_config=build_config
        )
        if build_config.system_libtreelite_dir:
            if not build_config.use_system_libtreelite:
                raise RuntimeError(
                    "use_system_libtreelite must be set to True if system_libtreelite_dir "
                    "is specified"
                )
            # Write path_config.py
            logger.info("Writing path_config.py...")
            custom_libpath = (
                pathlib.Path(build_config.system_libtreelite_dir).expanduser().resolve()
            )
            with open(pkg_path / "path_config.py", "w", encoding="utf-8") as f:
                print(
                    f"def get_custom_libpath():\n    return '{str(custom_libpath)}'",
                    file=f,
                )
        if not build_config.use_system_libtreelite:
            copy_with_logging(libtreelite, lib_path, logger=logger)

        with cd(workspace):
            wheel_name = hatchling.build.build_wheel(
                wheel_directory, config_settings, metadata_directory
            )
    return wheel_name


def build_sdist(
    sdist_directory: str,
    config_settings: Optional[Dict[str, Any]] = None,
) -> str:
    """Build a source distribution"""
    logger = logging.getLogger("treelite.packager.build_sdist")

    if config_settings:
        raise NotImplementedError(
            "Treelite's custom build backend doesn't support config_settings option "
            f"when building sdist. {config_settings=}"
        )

    cpp_src_dir = TOPLEVEL_DIR.parent
    if not cpp_src_dir.joinpath("CMakeLists.txt").exists():
        raise RuntimeError(f"Did not find CMakeLists.txt from {cpp_src_dir}")

    # Create tempdir with Python package + C++ sources
    with tempfile.TemporaryDirectory() as td:
        td_path = pathlib.Path(td)

        workspace = td_path / "sdist_workspace"
        workspace.mkdir()
        logger.info("Copying project files to temporary directory %s", str(workspace))

        copy_with_logging(TOPLEVEL_DIR / "pyproject.toml", workspace, logger=logger)
        copy_with_logging(TOPLEVEL_DIR / "hatch_build.py", workspace, logger=logger)
        copy_with_logging(TOPLEVEL_DIR / "README.rst", workspace, logger=logger)

        copytree_with_logging(
            TOPLEVEL_DIR / "treelite", workspace / "treelite", logger=logger
        )
        copytree_with_logging(
            TOPLEVEL_DIR / "packager", workspace / "packager", logger=logger
        )

        temp_cpp_src_dir = workspace / "cpp_src"
        copy_cpp_src_tree(cpp_src_dir, target_dir=temp_cpp_src_dir, logger=logger)

        with cd(workspace):
            sdist_name = hatchling.build.build_sdist(sdist_directory, config_settings)
    return sdist_name


def build_editable(
    wheel_directory: str,
    config_settings: Optional[Dict[str, Any]] = None,
    metadata_directory: Optional[str] = None,
) -> str:
    """Build an editable installation. We mostly delegate to Hatchling."""
    logger = logging.getLogger("treelite.packager.build_editable")

    if config_settings:
        raise NotImplementedError(
            "Treelite's custom build backend doesn't support config_settings option "
            f"when building editable installation. {config_settings=}"
        )

    if locate_local_libtreelite(TOPLEVEL_DIR, logger=logger) is None:
        raise RuntimeError(
            "To use the editable installation, first build libtreelite with CMake. "
            "See https://treelite.readthedocs.io/en/latest/install.html for detailed instructions."
        )

    return hatchling.build.build_editable(
        wheel_directory, config_settings, metadata_directory
    )



"""
Miscellaneous utilities
"""

import ctypes

import numpy as np

_CTYPES_TYPE_TABLE = {
    "uint32": ctypes.c_uint32,
    "float32": ctypes.c_float,
    "float64": ctypes.c_double,
}


_NUMPY_TYPE_TABLE = {"uint32": np.uint32, "float32": np.float32, "float64": np.float64}


def typestr_to_ctypes_type(type_info):
    """Obtain ctypes type corresponding to a given Type str"""
    return _CTYPES_TYPE_TABLE[type_info]


def typestr_to_numpy_type(type_info):
    """Obtain ctypes type corresponding to a given Type str"""
    return _NUMPY_TYPE_TABLE[type_info]


def c_str(string):
    """Convert a Python string to C string"""
    return ctypes.c_char_p(string.encode("utf-8"))


def py_str(string):
    """Convert C string back to Python string"""
    return string.decode("utf-8")


def c_array(ctype, values):
    """
    Convert a Python byte array to C array

    WARNING
    -------
    DO NOT USE THIS FUNCTION if performance is critical. Instead, use np.array(*)
    with dtype option to explicitly convert type and then use
    ndarray.ctypes.data_as(*) to expose underlying buffer as C pointer.
    """
    return (ctype * len(values))(*values)



"""Find the path to Treelite dynamic library files."""

import os
import pathlib
import sys
from typing import List

from .path_config import get_custom_libpath


class TreeliteLibraryNotFound(Exception):
    """Error thrown by when Treelite is not found"""


def find_lib_path() -> List[pathlib.Path]:
    """Find the path to Treelite dynamic library files.

    Returns
    -------
    lib_path
       List of all found library path to Treelite
    """
    curr_path = pathlib.Path(__file__).expanduser().absolute().parent
    dll_path = [
        # When installed, libtreelite will be installed in <site-package-dir>/lib
        curr_path / "lib",
        # Editable installation
        curr_path.parent.parent / "build",
        # Use libtreelite from a system prefix, if available. This should be the last option.
        pathlib.Path(sys.base_prefix).expanduser().resolve() / "lib",
    ]
    custom_libpath = get_custom_libpath()  # pylint: disable=assignment-from-none
    if custom_libpath:
        dll_path.insert(0, pathlib.Path(custom_libpath).expanduser().resolve())

    if sys.platform == "win32":
        # On Windows, Conda may install libs in different paths
        sys_prefix = pathlib.Path(sys.base_prefix)
        dll_path.extend(
            [
                sys_prefix / "bin",
                sys_prefix / "Library",
                sys_prefix / "Library" / "bin",
                sys_prefix / "Library" / "lib",
            ]
        )
        dll_path = [p.joinpath("treelite.dll") for p in dll_path]
    elif sys.platform.startswith(("linux", "freebsd", "emscripten", "OS400")):
        dll_path = [p.joinpath("libtreelite.so") for p in dll_path]
    elif sys.platform == "darwin":
        dll_path = [p.joinpath("libtreelite.dylib") for p in dll_path]
    elif sys.platform == "cygwin":
        dll_path = [p.joinpath("cygtreelite.dll") for p in dll_path]
    else:
        raise RuntimeError(f"Unrecognized platform: {sys.platform}")

    lib_path = [p for p in dll_path if p.exists() and p.is_file()]

    # TREELITE_BUILD_DOC is defined by sphinx conf.
    if not lib_path and not os.environ.get("TREELITE_BUILD_DOC", False):
        link = "https://treelite.readthedocs.io/en/latest/install.html"
        msg = (
            "Cannot find Treelite Library in the candidate path.  "
            + "List of candidates:\n- "
            + ("\n- ".join(str(x) for x in dll_path))
            + "\nTreelite Python package path: "
            + str(curr_path)
            + "\nsys.base_prefix: "
            + sys.base_prefix
            + "\nSee: "
            + link
            + " for installing Treelite."
        )
        raise TreeliteLibraryNotFound(msg)
    return lib_path



"""Interface with native lib"""

import ctypes
import os
import sys
import warnings

from .libpath import TreeliteLibraryNotFound, find_lib_path
from .util import py_str


class TreeliteError(Exception):
    """Error thrown by Treelite"""


@ctypes.CFUNCTYPE(None, ctypes.c_char_p)
def _log_callback(msg: bytes) -> None:
    """Redirect logs from native library into Python console"""
    print(py_str(msg))


@ctypes.CFUNCTYPE(None, ctypes.c_char_p)
def _warn_callback(msg: bytes) -> None:
    """Redirect warnings from native library into Python console"""
    warnings.warn(py_str(msg))


def _load_lib():
    """Load Treelite Library."""
    lib_path = [str(x) for x in find_lib_path()]
    if not lib_path:
        # Building docs
        return None  # type: ignore
    if sys.version_info >= (3, 8) and sys.platform == "win32":
        # pylint: disable=no-member
        os.add_dll_directory(
            os.path.join(os.path.normpath(sys.base_prefix), "Library", "bin")
        )
    lib = ctypes.CDLL(lib_path[0], mode=ctypes.RTLD_GLOBAL)
    lib.TreeliteGetLastError.restype = ctypes.c_char_p
    lib.log_callback = _log_callback
    lib.warn_callback = _warn_callback
    if lib.TreeliteRegisterLogCallback(lib.log_callback) != 0:
        raise TreeliteError(py_str(lib.TreeliteGetLastError()))
    if lib.TreeliteRegisterWarningCallback(lib.warn_callback) != 0:
        raise TreeliteError(py_str(lib.TreeliteGetLastError()))
    return lib


# Load the Treelite library globally
# (do not load if called by Sphinx)
if "sphinx" in sys.modules:
    try:
        _LIB = _load_lib()
    except TreeliteLibraryNotFound:
        _LIB = None
else:
    _LIB = _load_lib()


def _check_call(ret: int) -> None:
    """Check the return value of C API call

    This function will raise exception when error occurs.
    Wrap every API call with this function.

    Parameters
    ----------
    ret :
        return value from API calls
    """
    if ret != 0:
        raise TreeliteError(_LIB.TreeliteGetLastError().decode("utf-8"))



"""
Utility functions for implementing PEP 517 backend
"""

import logging
import pathlib
import shutil


def copytree_with_logging(
    src: pathlib.Path, dest: pathlib.Path, logger: logging.Logger
) -> None:
    """Call shutil.copytree() with logging"""
    logger.info("Copying %s -> %s", str(src), str(dest))
    shutil.copytree(src, dest)


def copy_with_logging(
    src: pathlib.Path, dest: pathlib.Path, logger: logging.Logger
) -> None:
    """Call shutil.copy() with logging"""
    if dest.is_dir():
        logger.info("Copying %s -> %s", str(src), str(dest / src.name))
    else:
        logger.info("Copying %s -> %s", str(src), str(dest))
    shutil.copy(src, dest)



"""Model loader ingest scikit-learn models into Treelite"""

from .importer import import_model


def import_model_with_model_builder(sklearn_model):
    """
    This function was removed in Treelite 4.0; please use :py:meth:`~treelite.sklearn.import_model`
    instead.
    """
    raise NotImplementedError(
        "treelite.sklearn.import_model_with_model_builder() was removed in Treelite 4.0. "
        "Please use treelite.sklearn.import_model() instead."
    )


__all__ = ["import_model", "import_model_with_model_builder"]



"""Treelite module"""

import pathlib

from . import frontend, gtil, model_builder, sklearn
from .core import TreeliteError
from .model import Model
from .model_builder_legacy import ModelBuilder

VERSION_FILE = pathlib.Path(__file__).parent / "VERSION"
with open(VERSION_FILE, "r", encoding="UTF-8") as _f:
    __version__ = _f.read().strip()

__all__ = [
    "Model",
    "frontend",
    "gtil",
    "sklearn",
    "model_builder",
    "TreeliteError",
    "ModelBuilder",
    "__version__",
]



"""Utility functions for loading IsolationForest models"""

import numpy as np
from scipy.special import psi


def harmonic(number):
    """Calculates the n-th harmonic number"""
    return psi(number + 1) + np.euler_gamma


def expected_depth(n_remainder):
    """Calculates the expected isolation depth for a remainder of uniform points"""
    if n_remainder <= 1:
        return 0.0
    if n_remainder == 2:
        return 1.0
    return float(2 * (harmonic(n_remainder) - 1))


def calculate_depths(isolation_depths, tree, curr_node, curr_depth):
    """Fill in an array of isolation depths for a scikit-learn isolation forest model"""
    if tree.children_left[curr_node] == -1:
        isolation_depths[curr_node] = curr_depth + expected_depth(
            tree.n_node_samples[curr_node]
        )
    else:
        calculate_depths(
            isolation_depths, tree, tree.children_left[curr_node], curr_depth + 1
        )
        calculate_depths(
            isolation_depths, tree, tree.children_right[curr_node], curr_depth + 1
        )



"""Compatibility layer to enable API migration"""

import ctypes
import json
from typing import Any, Optional, Union

from packaging.version import parse as parse_version

from .core import _LIB, TreeliteError, _check_call
from .util import c_str


def load_xgboost_model_legacy_binary(filename: str) -> Any:
    """
    Load a tree ensemble model from XGBoost model, stored using
    the legacy binary format.

    TODO(hcho3): Move the implementation to treelite.frontend once
                 Model.load() is removed.
    """
    handle = ctypes.c_void_p()
    _check_call(
        _LIB.TreeliteLoadXGBoostModelLegacyBinary(
            c_str(filename), c_str("{}"), ctypes.byref(handle)
        )
    )
    return handle


def load_xgboost_model(filename: str, *, allow_unknown_field: bool) -> Any:
    """
    Load a tree ensemble model from XGBoost model, stored using the JSON format.

    TODO(hcho3): Move the implementation to treelite.frontend once
                 Model.load() is removed.
    """
    parser_config = {"allow_unknown_field": allow_unknown_field}
    parser_config_str = json.dumps(parser_config)
    handle = ctypes.c_void_p()
    _check_call(
        _LIB.TreeliteLoadXGBoostModel(
            c_str(filename), c_str(parser_config_str), ctypes.byref(handle)
        )
    )
    return handle


def load_lightgbm_model(filename: str) -> Any:
    """
    Load a tree ensemble model from a LightGBM model file

    TODO(hcho3): Move the implementation to treelite.frontend once
                 Model.load() is removed.
    """
    handle = ctypes.c_void_p()
    _check_call(
        _LIB.TreeliteLoadLightGBMModel(
            c_str(filename), c_str("{}"), ctypes.byref(handle)
        )
    )
    return handle


def from_xgboost_json(
    model_json_str: Union[str, bytes, bytearray],
    *,
    allow_unknown_field: Optional[bool] = False,
) -> Any:
    """
    Load a tree ensemble model from a string containing XGBoost JSON.

    TODO(hcho3): Move the implementation to treelite.frontend once
                 Model.from_xgboost_json() is removed.
    """
    parser_config = {"allow_unknown_field": allow_unknown_field}
    parser_config_str = json.dumps(parser_config)

    handle = ctypes.c_void_p()
    length = len(model_json_str)
    if isinstance(model_json_str, (bytes, bytearray)):
        json_buffer = ctypes.create_string_buffer(bytes(model_json_str), length)
        _check_call(
            _LIB.TreeliteLoadXGBoostModelFromString(
                json_buffer,
                ctypes.c_size_t(length),
                c_str(parser_config_str),
                ctypes.byref(handle),
            )
        )
    else:
        _check_call(
            _LIB.TreeliteLoadXGBoostModelFromString(
                c_str(model_json_str),
                ctypes.c_size_t(length),
                c_str(parser_config_str),
                ctypes.byref(handle),
            )
        )
    return handle


def from_xgboost(booster: Any) -> Any:
    """
    Load a tree ensemble model from an XGBoost Booster object.

    TODO(hcho3): Move the implementation to treelite.frontend once
                 Model.from_xgboost() is removed.
    """
    try:
        import xgboost
    except ImportError as e:
        raise TreeliteError(
            "xgboost module must be installed to read from "
            + "`xgboost.Booster` object"
        ) from e
    if not isinstance(booster, xgboost.Booster):
        raise ValueError("booster must be of type `xgboost.Booster`")
    xgb_version = parse_version(xgboost.__version__)
    if xgb_version > parse_version("1.5.2"):
        # For XGBoost version 1.6.0 and later, use save_raw() to export models as JSON string
        model_json_str = booster.save_raw(raw_format="json")
        return from_xgboost_json(model_json_str)
    if xgb_version >= parse_version("1.0.0"):
        # Prior to version 1.6.0, XGBoost offer a method to export models as JSON string
        # in-memory. So use __getstate__ instead.
        model_json_str = booster.__getstate__()["handle"]
        return from_xgboost_json(model_json_str)
    # If pre-1.0.0 version of XGBoost is used, use legacy serialization
    handle = ctypes.c_void_p()
    buffer = booster.save_raw()
    ptr = (ctypes.c_char * len(buffer)).from_buffer(buffer)
    length = ctypes.c_size_t(len(buffer))
    _check_call(
        _LIB.TreeliteLoadXGBoostModelLegacyBinaryFromMemoryBuffer(
            ptr, length, ctypes.byref(handle)
        )
    )
    return handle


def from_lightgbm(booster: Any) -> Any:
    """
    Load a tree ensemble model from a LightGBM Booster object

    TODO(hcho3): Move the implementation to treelite.frontend once
                 Model.from_lightgbm() is removed.
    """
    handle = ctypes.c_void_p()
    # Attempt to import lightgbm
    try:
        import lightgbm
    except ImportError as e:
        raise TreeliteError(
            "lightgbm module must be installed to read from `lightgbm.Booster` object"
        ) from e
    if not isinstance(booster, lightgbm.Booster):
        raise ValueError("booster must be of type `lightgbm.Booster`")
    model_str = booster.model_to_string()
    _check_call(
        _LIB.TreeliteLoadLightGBMModelFromString(
            c_str(model_str), c_str("{}"), ctypes.byref(handle)
        )
    )
    return handle



"""Converter to ingest scikit-learn models into Treelite"""

import ctypes

import numpy as np
from packaging.version import parse as parse_version

from ..core import _LIB, TreeliteError, _check_call
from ..frontend import Model
from ..util import c_array
from .isolation_forest import calculate_depths, expected_depth


class ArrayOfArrays:
    """
    Utility class to marshall a collection of arrays in order to pass to a C function
    """

    def __init__(self, *, dtype):
        int8_ptr_type = ctypes.POINTER(ctypes.c_int8)
        int64_ptr_type = ctypes.POINTER(ctypes.c_int64)
        uint32_ptr_type = ctypes.POINTER(ctypes.c_uint32)
        float64_ptr_type = ctypes.POINTER(ctypes.c_double)
        void_ptr_type = ctypes.c_void_p
        if dtype == np.int64:
            self.ptr_type = int64_ptr_type
        elif dtype == np.float64:
            self.ptr_type = float64_ptr_type
        elif dtype == np.uint32:
            self.ptr_type = uint32_ptr_type
        elif dtype == np.int8:
            self.ptr_type = int8_ptr_type
        elif dtype == "void":
            self.ptr_type = void_ptr_type
        else:
            raise ValueError(f"dtype {dtype} is not supported")
        self.dtype = dtype
        self.collection = []
        self.collection_ptr = []

    def add(self, array, *, expected_shape=None):
        """Add an array to the collection"""
        if self.dtype != "void":
            assert array.dtype == self.dtype
        if expected_shape:
            assert (
                array.shape == expected_shape
            ), f"Expected shape: {expected_shape}, Got shape {array.shape}"
        v = np.asarray(array, dtype=self.dtype, order="C")
        self.collection.append(v)

    def as_c_array(self):
        """Prepare the collection to pass as an argument of a C function"""
        if not self.collection:
            return None  # return nullptr to represent empty array
        for v in self.collection:
            self.collection_ptr.append(v.ctypes.data_as(self.ptr_type))
        return c_array(self.ptr_type, self.collection_ptr)


def import_model(sklearn_model):
    # pylint: disable=R0914,R0912,R0915
    """
    Load a tree ensemble model from a scikit-learn model object

    Note
    ----
    For :py:class:`~sklearn.ensemble.IsolationForest`, the loaded model will calculate the outlier
    score using the standardized ratio as proposed in the original reference,
    which matches with :py:meth:`~sklearn.ensemble.IsolationForest._compute_chunked_score_samples`
    but is a bit different from :py:meth:`~sklearn.ensemble.IsolationForest.decision_function`.

    Parameters
    ----------
    sklearn_model : object of type \
                    :py:class:`~sklearn.ensemble.RandomForestRegressor` / \
                    :py:class:`~sklearn.ensemble.RandomForestClassifier` / \
                    :py:class:`~sklearn.ensemble.ExtraTreesRegressor` / \
                    :py:class:`~sklearn.ensemble.ExtraTreesClassifier` / \
                    :py:class:`~sklearn.ensemble.GradientBoostingRegressor` / \
                    :py:class:`~sklearn.ensemble.GradientBoostingClassifier` / \
                    :py:class:`~sklearn.ensemble.HistGradientBoostingRegressor` / \
                    :py:class:`~sklearn.ensemble.HistGradientBoostingClassifier` / \
                    :py:class:`~sklearn.ensemble.IsolationForest`
        Python handle to scikit-learn model

    Returns
    -------
    model : :py:class:`Model`
        Loaded model

    Example
    -------

    .. code-block:: python
      :emphasize-lines: 8

      import sklearn.datasets
      import sklearn.ensemble
      X, y = sklearn.datasets.load_diabetes(return_X_y=True)
      clf = sklearn.ensemble.RandomForestRegressor(n_estimators=10)
      clf.fit(X, y)

      import treelite.sklearn
      model = treelite.sklearn.import_model(clf)

    Note
    ----
    This function does not yet support categorical splits in
    :py:class:`~sklearn.ensemble.HistGradientBoostingRegressor` and
    :py:class:`~sklearn.ensemble.HistGradientBoostingClassifier`.
    If you are using either estimator types, make sure that all
    test nodes have numerical test conditions.
    """
    try:
        from sklearn.dummy import DummyClassifier, DummyRegressor
        from sklearn.ensemble import ExtraTreesClassifier as ExtraTreesC
        from sklearn.ensemble import ExtraTreesRegressor as ExtraTreesR
        from sklearn.ensemble import GradientBoostingClassifier as GradientBoostingC
        from sklearn.ensemble import GradientBoostingRegressor as GradientBoostingR
        from sklearn.ensemble import (
            HistGradientBoostingClassifier as HistGradientBoostingC,
        )
        from sklearn.ensemble import (
            HistGradientBoostingRegressor as HistGradientBoostingR,
        )
        from sklearn.ensemble import IsolationForest
        from sklearn.ensemble import RandomForestClassifier as RandomForestC
        from sklearn.ensemble import RandomForestRegressor as RandomForestR
    except ImportError as e:
        raise TreeliteError("This function requires scikit-learn package") from e

    if isinstance(sklearn_model, (HistGradientBoostingR, HistGradientBoostingC)):
        return _import_hist_gradient_boosting(sklearn_model)

    if isinstance(sklearn_model, (RandomForestR, ExtraTreesR)):
        # pylint: disable=C3001
        leaf_value_expected_shape = lambda node_count: (  # noqa: E731
            node_count,
            sklearn_model.n_outputs_,
            1,
        )
    elif isinstance(sklearn_model, (RandomForestC, ExtraTreesC)):
        # pylint: disable=C3001
        try:  # multi-target
            max_n_classes = max(sklearn_model.n_classes_)
        except TypeError:  # single-target
            max_n_classes = sklearn_model.n_classes_
        leaf_value_expected_shape = lambda node_count: (  # noqa: E731
            node_count,
            sklearn_model.n_outputs_,
            max_n_classes,
        )
    elif isinstance(
        sklearn_model, (GradientBoostingR, GradientBoostingC, IsolationForest)
    ):
        # pylint: disable=C3001
        leaf_value_expected_shape = lambda node_count: (  # noqa: E731
            node_count,
            1,
            1,
        )
    else:
        raise TreeliteError(
            f"Not supported model type: {sklearn_model.__class__.__name__}"
        )

    if isinstance(sklearn_model, IsolationForest):
        ratio_c = expected_depth(sklearn_model.max_samples_)

    node_count = []
    children_left = ArrayOfArrays(dtype=np.int64)
    children_right = ArrayOfArrays(dtype=np.int64)
    feature = ArrayOfArrays(dtype=np.int64)
    threshold = ArrayOfArrays(dtype=np.float64)
    value = ArrayOfArrays(dtype=np.float64)
    n_node_samples = ArrayOfArrays(dtype=np.int64)
    weighted_n_node_samples = ArrayOfArrays(dtype=np.float64)
    impurity = ArrayOfArrays(dtype=np.float64)
    for estimator in sklearn_model.estimators_:
        if isinstance(sklearn_model, (GradientBoostingR, GradientBoostingC)):
            estimator_range = estimator
            learning_rate = sklearn_model.learning_rate
        else:
            estimator_range = [estimator]
            learning_rate = 1.0
        if isinstance(sklearn_model, IsolationForest):
            isolation_depths = np.zeros(
                estimator.tree_.n_node_samples.shape[0], dtype="float64"
            )
            calculate_depths(isolation_depths, estimator.tree_, 0, 0.0)
        for sub_estimator in estimator_range:
            tree = sub_estimator.tree_
            node_count.append(tree.node_count)
            children_left.add(tree.children_left, expected_shape=(tree.node_count,))
            children_right.add(tree.children_right, expected_shape=(tree.node_count,))
            feature.add(tree.feature, expected_shape=(tree.node_count,))
            threshold.add(tree.threshold, expected_shape=(tree.node_count,))
            if isinstance(sklearn_model, IsolationForest):
                value.add(
                    isolation_depths.reshape((-1, 1, 1)),
                    expected_shape=leaf_value_expected_shape(tree.node_count),
                )
            else:
                # Note: for gradient boosted trees, we shrink each leaf output by the
                # learning rate
                value.add(
                    tree.value * learning_rate,
                    expected_shape=leaf_value_expected_shape(tree.node_count),
                )
            n_node_samples.add(tree.n_node_samples, expected_shape=(tree.node_count,))
            weighted_n_node_samples.add(
                tree.weighted_n_node_samples, expected_shape=(tree.node_count,)
            )
            impurity.add(tree.impurity, expected_shape=(tree.node_count,))

    handle = ctypes.c_void_p()
    if isinstance(sklearn_model, (RandomForestR, ExtraTreesR)):
        _check_call(
            _LIB.TreeliteLoadSKLearnRandomForestRegressor(
                ctypes.c_int(sklearn_model.n_estimators),
                ctypes.c_int(sklearn_model.n_features_in_),
                ctypes.c_int(sklearn_model.n_outputs_),
                c_array(ctypes.c_int64, node_count),
                children_left.as_c_array(),
                children_right.as_c_array(),
                feature.as_c_array(),
                threshold.as_c_array(),
                value.as_c_array(),
                n_node_samples.as_c_array(),
                weighted_n_node_samples.as_c_array(),
                impurity.as_c_array(),
                ctypes.byref(handle),
            )
        )
    elif isinstance(sklearn_model, IsolationForest):
        _check_call(
            _LIB.TreeliteLoadSKLearnIsolationForest(
                ctypes.c_int(sklearn_model.n_estimators),
                ctypes.c_int(sklearn_model.n_features_in_),
                c_array(ctypes.c_int64, node_count),
                children_left.as_c_array(),
                children_right.as_c_array(),
                feature.as_c_array(),
                threshold.as_c_array(),
                value.as_c_array(),
                n_node_samples.as_c_array(),
                weighted_n_node_samples.as_c_array(),
                impurity.as_c_array(),
                ctypes.c_double(ratio_c),
                ctypes.byref(handle),
            )
        )
    elif isinstance(sklearn_model, (RandomForestC, ExtraTreesC)):
        n_classes = np.array(sklearn_model.n_classes_, dtype=np.int32)
        _check_call(
            _LIB.TreeliteLoadSKLearnRandomForestClassifier(
                ctypes.c_int(sklearn_model.n_estimators),
                ctypes.c_int(sklearn_model.n_features_in_),
                ctypes.c_int(sklearn_model.n_outputs_),
                n_classes.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)),
                c_array(ctypes.c_int64, node_count),
                children_left.as_c_array(),
                children_right.as_c_array(),
                feature.as_c_array(),
                threshold.as_c_array(),
                value.as_c_array(),
                n_node_samples.as_c_array(),
                weighted_n_node_samples.as_c_array(),
                impurity.as_c_array(),
                ctypes.byref(handle),
            )
        )
    elif isinstance(sklearn_model, GradientBoostingR):
        if sklearn_model.init_ == "zero":
            base_scores = np.array([0], dtype=np.float64)
        elif isinstance(sklearn_model.init_, (DummyRegressor,)):
            base_scores = np.array(sklearn_model.init_.constant_, dtype=np.float64)
            assert base_scores.size == 1
        else:
            raise NotImplementedError("Custom init estimator not supported")
        _check_call(
            _LIB.TreeliteLoadSKLearnGradientBoostingRegressor(
                ctypes.c_int(sklearn_model.n_estimators),
                ctypes.c_int(sklearn_model.n_features_in_),
                c_array(ctypes.c_int64, node_count),
                children_left.as_c_array(),
                children_right.as_c_array(),
                feature.as_c_array(),
                threshold.as_c_array(),
                value.as_c_array(),
                n_node_samples.as_c_array(),
                weighted_n_node_samples.as_c_array(),
                impurity.as_c_array(),
                base_scores.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                ctypes.byref(handle),
            )
        )
    elif isinstance(sklearn_model, GradientBoostingC):
        if sklearn_model.init_ == "zero":
            base_scores = np.zeros(shape=(sklearn_model.n_classes_,), dtype=np.float64)
        elif isinstance(sklearn_model.init_, (DummyClassifier,)):
            if sklearn_model.init_.strategy != "prior":
                raise NotImplementedError("Custom init estimator not supported")
            # pylint: disable=W0212
            base_scores = np.array(
                sklearn_model._raw_predict_init(
                    np.zeros((1, sklearn_model.n_features_in_))
                ),
                dtype=np.float64,
            )
        else:
            raise NotImplementedError("Custom init estimator not supported")
        _check_call(
            _LIB.TreeliteLoadSKLearnGradientBoostingClassifier(
                ctypes.c_int(sklearn_model.n_estimators),
                ctypes.c_int(sklearn_model.n_features_in_),
                ctypes.c_int(sklearn_model.n_classes_),
                c_array(ctypes.c_int64, node_count),
                children_left.as_c_array(),
                children_right.as_c_array(),
                feature.as_c_array(),
                threshold.as_c_array(),
                value.as_c_array(),
                n_node_samples.as_c_array(),
                weighted_n_node_samples.as_c_array(),
                impurity.as_c_array(),
                base_scores.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                ctypes.byref(handle),
            )
        )
    else:
        raise TreeliteError(
            f"Unsupported model type {sklearn_model.__class__.__name__}: "
            + "currently random forests, extremely randomized trees, and gradient "
            + "boosted trees are supported"
        )
    return Model(handle=handle)


def _import_hist_gradient_boosting(sklearn_model):
    # pylint: disable=R0914,W0212,too-many-branches
    """Load HistGradientBoostingClassifier / HistGradientBoostingRegressor"""
    from sklearn.ensemble import HistGradientBoostingClassifier as HistGradientBoostingC
    from sklearn.ensemble import HistGradientBoostingRegressor as HistGradientBoostingR

    # Arrays to be passed to C API functions
    (
        known_cat_bitsets,
        f_idx_map,
    ) = sklearn_model._bin_mapper.make_known_categories_bitsets()

    cat_remapper = ArrayOfArrays(dtype=np.int64)

    from sklearn import __version__ as sklearn_version

    if (
        parse_version(sklearn_version) >= parse_version("1.4.0")
        and getattr(sklearn_model, "_preprocessor", None) is not None
    ):
        # Ensure that the model's internals did not change
        assert len(sklearn_model._preprocessor.transformers_) == 2
        assert len(sklearn_model._preprocessor.transformers_[0]) == 3
        assert len(sklearn_model._preprocessor.transformers_[1]) == 3
        assert sklearn_model._preprocessor.transformers_[0][0] == "encoder"
        assert sklearn_model._preprocessor.transformers_[1][0] == "numerical"

        for cats in sklearn_model._preprocessor.transformers_[0][1].categories_:
            if cats.dtype.type is np.str_:
                raise NotImplementedError("String categories are not supported")
            cats = cats.astype(np.int64)
            cat_remapper.add(cats)

        feat_remapper = np.zeros((sklearn_model.n_features_in_,), dtype=np.int32)
        n_categorical = sklearn_model.is_categorical_.sum()
        cat_idx, num_idx = 0, 0
        for i, e in enumerate(sklearn_model.is_categorical_):
            if e:
                feat_remapper[cat_idx] = i
                cat_idx += 1
            else:
                feat_remapper[n_categorical + num_idx] = i
                num_idx += 1
    else:
        feat_remapper = np.arange(
            start=0, stop=sklearn_model.n_features_in_, dtype=np.int32
        )

    n_categorical_splits = known_cat_bitsets.shape[0]
    n_trees = 0
    nodes = ArrayOfArrays(dtype="void")
    raw_left_cat_bitsets = ArrayOfArrays(dtype=np.uint32)
    node_count = []
    itemsize = None

    for estimator in sklearn_model._predictors:
        estimator_range = estimator
        for sub_estimator in estimator_range:
            nodes.add(sub_estimator.nodes)
            raw_left_cat_bitsets.add(sub_estimator.raw_left_cat_bitsets)
            node_count.append(len(sub_estimator.nodes))
            n_trees += 1
            if itemsize is None:
                itemsize = sub_estimator.nodes.itemsize
            elif itemsize != sub_estimator.nodes.itemsize:
                raise RuntimeError("itemsize mismatch")

    handle = ctypes.c_void_p()
    if isinstance(sklearn_model, (HistGradientBoostingR,)):
        _check_call(
            _LIB.TreeliteLoadSKLearnHistGradientBoostingRegressor(
                ctypes.c_int(sklearn_model.n_iter_),
                ctypes.c_int(sklearn_model.n_features_in_),
                c_array(ctypes.c_int64, node_count),
                nodes.as_c_array(),
                ctypes.c_int(itemsize),
                ctypes.c_int32(n_categorical_splits),
                raw_left_cat_bitsets.as_c_array(),
                known_cat_bitsets.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                f_idx_map.ctypes.data_as(ctypes.POINTER(ctypes.c_uint32)),
                feat_remapper.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)),
                cat_remapper.as_c_array(),
                sklearn_model._baseline_prediction.ctypes.data_as(
                    ctypes.POINTER(ctypes.c_double)
                ),
                ctypes.byref(handle),
            )
        )
    elif isinstance(sklearn_model, (HistGradientBoostingC,)):
        _check_call(
            _LIB.TreeliteLoadSKLearnHistGradientBoostingClassifier(
                ctypes.c_int(sklearn_model.n_iter_),
                ctypes.c_int(sklearn_model.n_features_in_),
                ctypes.c_int(len(sklearn_model.classes_)),
                c_array(ctypes.c_int64, node_count),
                nodes.as_c_array(),
                ctypes.c_int(itemsize),
                ctypes.c_int32(n_categorical_splits),
                raw_left_cat_bitsets.as_c_array(),
                known_cat_bitsets.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                f_idx_map.ctypes.data_as(ctypes.POINTER(ctypes.c_uint32)),
                feat_remapper.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)),
                cat_remapper.as_c_array(),
                sklearn_model._baseline_prediction.ctypes.data_as(
                    ctypes.POINTER(ctypes.c_double)
                ),
                ctypes.byref(handle),
            )
        )
    else:
        raise TreeliteError(
            f"Unsupported model type {sklearn_model.__class__.__name__}"
        )
    return Model(handle=handle)



"""
General Tree Inference Library (GTIL)
"""

import ctypes
import json
from dataclasses import asdict, dataclass
from typing import Union

import numpy as np
from scipy.sparse import csr_matrix

from ..core import _LIB, _check_call
from ..frontend import Model
from ..util import c_str, typestr_to_ctypes_type, typestr_to_numpy_type


@dataclass
class GTILConfig:
    """Object holding configuration data"""

    predict_type: str
    # Prediction type is one of the following:
    # * "default": Usual prediction method. Sum over trees and apply post-processing.
    # * "raw": Sum over trees, but don't apply post-processing; get raw margin scores
    #          instead.
    # * "leaf_id": Output one (integer) leaf ID per tree.
    # * "score_per_tree": Output one or more margin scores per tree.
    nthread: int = -1
    # <= 0 indicates using all threads

    def to_json(self):
        """Convert configuration object into a JSON string"""
        return json.dumps(asdict(self))

    def __post_init__(self):
        self._handle = ctypes.c_void_p()
        _check_call(
            _LIB.TreeliteGTILParseConfig(
                c_str(self.to_json()), ctypes.byref(self._handle)
            )
        )

    def __del__(self):
        if self._handle is not None:
            _check_call(_LIB.TreeliteGTILDeleteConfig(self._handle))

    @property
    def handle(self):
        """Access the handle to the associated C++ object"""
        return self._handle


def predict(
    model: Model,
    data: Union[np.ndarray, csr_matrix],
    *,
    nthread: int = -1,
    pred_margin: bool = False,
):
    """
    Predict with a Treelite model using the General Tree Inference Library (GTIL).

    Parameters
    ----------
    model : :py:class:`Model` object
        Treelite model object
    data : :py:class:`numpy.ndarray` / :py:class:`scipy.sparse.csr_matrix`
        Data matrix, with which to run prediction
    nthread : :py:class:`int <python:int>`, optional
        Number of CPU cores to use in prediction. If <= 0, use all CPU cores.
    pred_margin : bool
        Whether to produce raw margin scores. If pred_margin=True, post-processing
        is no longer applied and raw margin scores are produced.

    Returns
    -------
    prediction : :py:class:`numpy.ndarray` array
        Prediction output. Expected dimensions: (num_row, num_target, max(num_class))
    """
    predict_type = "raw" if pred_margin else "default"

    config = GTILConfig(nthread=nthread, predict_type=predict_type)
    return _predict_impl(model, data, config=config)


def predict_leaf(
    model: Model,
    data: Union[np.ndarray, csr_matrix],
    *,
    nthread: int = -1,
):
    """
    Predict with a Treelite model, outputting the leaf node's ID for each row.

    Parameters
    ----------
    model : :py:class:`Model` object
        Treelite model object
    data : :py:class:`numpy.ndarray` / :py:class:`scipy.sparse.csr_matrix`
        Data matrix, with which to run prediction
    nthread : :py:class:`int <python:int>`, optional
        Number of CPU cores to use in prediction. If <= 0, use all CPU cores.

    Returns
    -------
    prediction : :py:class:`numpy.ndarray` array
        Prediction output. Expected output dimensions: (num_row, num_tree)

    Notes
    -----
    Treelite assigns a unique integer ID for every node in the tree, including leaf
    nodes as well as internal nodes. It does so by traversing the tree breadth-first.
    So, for example, the root node is assigned ID 0, and the two nodes at depth=1 is
    assigned ID 1 and 2, respectively.
    Call :py:meth:`treelite.Model.dump_as_json` to obtain the ID of every tree node.
    """

    config = GTILConfig(nthread=nthread, predict_type="leaf_id")
    return _predict_impl(model, data, config=config)


def predict_per_tree(
    model: Model, data: Union[np.ndarray, csr_matrix], *, nthread: int = -1
):
    """
    Predict with a Treelite model and output prediction of each tree.
    This function computes one or more margin scores per tree.

    Parameters
    ----------
    model : :py:class:`Model` object
        Treelite model object
    data : :py:class:`numpy.ndarray` / :py:class:`scipy.sparse.csr_matrix`
        Data matrix, with which to run prediction
    nthread : :py:class:`int <python:int>`, optional
        Number of CPU cores to use in prediction. If <= 0, use all CPU cores.

    Returns
    -------
    prediction : :py:class:`numpy.ndarray` array
        Prediction output. Expected output dimensions:
        (num_row, num_tree, leaf_vector_shape[0] * leaf_vector_shape[1])
    """

    config = GTILConfig(nthread=nthread, predict_type="score_per_tree")
    return _predict_impl(model, data, config=config)


def _predict_impl(
    model: Model, data: Union[np.ndarray, csr_matrix], *, config: GTILConfig
) -> np.ndarray:
    # Validate parameters
    if not isinstance(model, Model):
        raise ValueError('Argument "model" must be a Model type')
    if not isinstance(data, (np.ndarray, csr_matrix)):
        raise ValueError(
            'Argument "data" must be either 2D NumPy array or Scipy CSR matrix'
        )
    if len(data.shape) != 2:
        raise ValueError('Argument "data" must be a 2D array')

    is_dense = isinstance(data, np.ndarray)

    if is_dense:
        data = np.asarray(
            data, dtype=typestr_to_numpy_type(model.input_type), order="C"
        )
        if data.shape[1] < model.num_feature:
            # Pad missing features with NAs
            data = np.pad(
                data,
                ((0, 0), (0, model.num_feature - data.shape[1])),
                "constant",
                constant_values="nan",
            )
            assert data.shape[1] == model.num_feature
    else:
        assert isinstance(data, csr_matrix)
        elems = np.asarray(
            data.data,
            dtype=typestr_to_numpy_type(model.input_type),
            order="C",
        )
        col_ind = np.asarray(data.indices, dtype=np.uint64, order="C")
        row_ptr = np.asarray(data.indptr, dtype=np.uint64, order="C")
    output_shape_ptr = ctypes.POINTER(ctypes.c_uint64)()
    output_ndim = ctypes.c_uint64()
    _check_call(
        _LIB.TreeliteGTILGetOutputShape(
            model.handle,
            ctypes.c_uint64(data.shape[0]),
            config.handle,
            ctypes.byref(output_shape_ptr),
            ctypes.byref(output_ndim),
        )
    )
    output_shape = np.ctypeslib.as_array(output_shape_ptr, shape=(output_ndim.value,))
    out_result = np.zeros(
        shape=output_shape, dtype=typestr_to_numpy_type(model.output_type), order="C"
    )

    if is_dense:
        _check_call(
            _LIB.TreeliteGTILPredict(
                model.handle,
                data.ctypes.data_as(
                    ctypes.POINTER(typestr_to_ctypes_type(model.input_type))
                ),
                c_str(model.input_type),
                ctypes.c_uint64(data.shape[0]),
                out_result.ctypes.data_as(
                    ctypes.POINTER(typestr_to_ctypes_type(model.output_type))
                ),
                config.handle,
            )
        )
    else:
        _check_call(
            _LIB.TreeliteGTILPredictSparse(
                model.handle,
                elems.ctypes.data_as(
                    ctypes.POINTER(typestr_to_ctypes_type(model.input_type))
                ),
                c_str(model.input_type),
                col_ind.ctypes.data_as(ctypes.POINTER(ctypes.c_uint64)),
                row_ptr.ctypes.data_as(ctypes.POINTER(ctypes.c_uint64)),
                ctypes.c_uint64(data.shape[0]),
                out_result.ctypes.data_as(
                    ctypes.POINTER(typestr_to_ctypes_type(model.output_type))
                ),
                config.handle,
            )
        )
    return out_result



"""Build configuration"""

import dataclasses
from typing import Any, Dict, List, Optional


@dataclasses.dataclass
class BuildConfiguration:  # pylint: disable=R0902
    """Configurations use when building libtreelite"""

    # Whether to enable OpenMP
    use_openmp: bool = True
    # Whether to hide C++ symbols
    hide_cxx_symbols: bool = True
    # Whether to use the Treelite library that's installed in the system prefix
    use_system_libtreelite: bool = False
    # Manually configure path for the Treelite library.
    # Only applicable when use_system_libtreelite=True
    system_libtreelite_dir: str = ""

    def _set_config_setting(self, config_settings: Dict[str, Any]) -> None:
        for field_name in config_settings:
            config_value = config_settings[field_name]
            if field_name == "system_libtreelite_dir":
                setattr(self, field_name, config_value)
            else:
                setattr(
                    self,
                    field_name,
                    (config_value.lower() in ["true", "1", "on"]),
                )

    def update(self, config_settings: Optional[Dict[str, Any]]) -> None:
        """Parse config_settings from Pip (or other PEP 517 frontend)"""
        if config_settings is not None:
            self._set_config_setting(config_settings)

    def get_cmake_args(self) -> List[str]:
        """Convert build configuration to CMake args"""
        cmake_args = []
        for field_name in [x.name for x in dataclasses.fields(self)]:
            if field_name in ["use_system_libtreelite", "system_libtreelite_dir"]:
                continue
            cmake_option = field_name.upper()
            cmake_value = "ON" if getattr(self, field_name) is True else "OFF"
            cmake_args.append(f"-D{cmake_option}={cmake_value}")
        return cmake_args



"""Functions to load and build model objects"""

from __future__ import annotations

import pathlib
from typing import Any, Union

from . import compat
from .model import Model


def load_xgboost_model_legacy_binary(filename: Union[str, pathlib.Path]) -> Model:
    """
    Load a tree ensemble model from XGBoost model, stored using
    the legacy binary format. Note: new XGBoost models should
    be stored in the JSON format, to take advantage of the
    latest functionalities of XGBoost.

    Parameters
    ----------
    filename :
        Path to model file

    Returns
    -------
    model : :py:class:`Model`
        Loaded model

    Example
    -------

    .. code-block:: python

       xgb_model = treelite.frontend.load_xgboost_model_legacy_binary(
           "xgboost_model.model")
    """
    return Model(handle=compat.load_xgboost_model_legacy_binary(str(filename)))


def load_xgboost_model(
    filename: Union[str, pathlib.Path], *, allow_unknown_field: bool = False
) -> Model:
    """
    Load a tree ensemble model from XGBoost model, stored using the JSON format.

    Parameters
    ----------
    filename :
        Path to model file
    allow_unknown_field:
        Whether to allow extra fields with unrecognized keys

    Returns
    -------
    model : :py:class:`Model`
        Loaded model

    Example
    -------

    .. code-block:: python

       xgb_model = treelite.frontend.load_xgboost_model("xgboost_model.json")
    """
    return Model(
        handle=compat.load_xgboost_model(
            str(filename), allow_unknown_field=allow_unknown_field
        )
    )


def load_lightgbm_model(filename: Union[str, pathlib.Path]) -> Model:
    """
    Load a tree ensemble model from a LightGBM model file.

    Parameters
    ----------
    filename :
        Path to model file

    Returns
    -------
    model : :py:class:`Model`
        Loaded model

    Example
    -------

    .. code-block:: python

       lgb_model = treelite.frontend.load_lightgbm_model("lightgbm_model.txt")
    """
    return Model(handle=compat.load_lightgbm_model(str(filename)))


def from_xgboost(booster: Any) -> Model:
    """
    Load a tree ensemble model from an XGBoost Booster object

    Parameters
    ----------
    booster : Object of type :py:class:`xgboost.Booster`
        Python handle to XGBoost model

    Returns
    -------
    model : :py:class:`Model`
        Loaded model
    """
    return Model(handle=compat.from_xgboost(booster))


def from_xgboost_json(
    model_json_str: Union[bytes, bytearray, str],
    *,
    allow_unknown_field: bool = False,
) -> Model:
    """
    Load a tree ensemble model from a string containing XGBoost JSON

    Parameters
    ----------
    model_json_str :
        A string specifying an XGBoost model in the XGBoost JSON format
    allow_unknown_field:
        Whether to allow extra fields with unrecognized keys

    Returns
    -------
    model: :py:class:`Model`
        Loaded model
    """
    return Model(
        handle=compat.from_xgboost_json(
            model_json_str, allow_unknown_field=allow_unknown_field
        )
    )


def from_lightgbm(booster: Any) -> Model:
    """
    Load a tree ensemble model from a LightGBM Booster object

    Parameters
    ----------
    booster : object of type :py:class:`lightgbm.Booster`
        Python handle to LightGBM model

    Returns
    -------
    model : :py:class:`Model`
        Loaded model
    """
    return Model(handle=compat.from_lightgbm(booster))


__all__ = [
    "load_xgboost_model_legacy_binary",
    "load_xgboost_model",
    "load_lightgbm_model",
    "from_xgboost",
    "from_xgboost_json",
    "from_lightgbm",
]
